;; Generated package description from spacious-padding.el  -*- no-byte-compile: t -*-
(define-package "spacious-padding" "0.3.0" "Increase the padding/spacing of frames and windows" '((emacs "28.1")) :commit "9d96d301d5bccf192daaf00dba64bca9979dcb5a" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou General Issues" . "~protesilaos/general-issues@lists.sr.ht") :keywords '("convenience" "focus" "writing" "presentation") :url "https://git.sr.ht/~protesilaos/spacious-padding")
